#ifndef _GNU_SOURCE
#define _GNU_SOURCE 1
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <error.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <limits.h>

#define AVAILDIR	"../conf.avail/"
#define CONFDIR		"/etc/fonts/conf.d/"
#define TEMPLATEDIR	"/usr/share/fontconfig/conf.avail/"

struct FCStruct {
	char* name;
	char* file;
};

static struct FCStruct Antialias[] = {
	{"enable", "10-yes-antialias.conf"},
	{"disable", "10-no-antialias.conf"},
	{"unknown", NULL},
	{NULL, NULL}
};

static struct FCStruct Autohint[] = {
	{"enable", "10-autohint.conf"},
	{"disable", NULL},
	{NULL, NULL}
};

static struct FCStruct Hinting[] = {
	{"enable", "10-hinting.conf"},
	{"disable", "10-unhinted.conf"},
	{"unknown", NULL},
	{NULL, NULL}
};

static struct FCStruct Hintstyle[] = {
	{"full", "10-hinting-full.conf"},
	{"slight", "10-hinting-slight.conf"},
	{"medium", "10-hinting-medium.conf"},
	{"none", "10-hinting-none.conf"},
	{"unknown", NULL},
	{NULL, NULL}
};

static struct FCStruct RGBA[] = {
	{"bgr", "10-sub-pixel-bgr.conf"},
	{"rgb", "10-sub-pixel-rgb.conf"},
	{"vbgr", "10-sub-pixel-vbgr.conf"},
	{"vrgb", "10-sub-pixel-vrgb.conf"},
	{"disable", "10-sub-pixel-none.conf"},
	{"unknown", NULL},
	{NULL, NULL}
};

static struct FCStruct LCDFilter[] = {
	{"default", "11-lcdfilter-default.conf"},
	{"legacy", "11-lcdfilter-legacy.conf"},
	{"light", "11-lcdfilter-light.conf"},
	{"none", "11-lcdfilter-none.conf"},
	{"unknown", NULL},
	{NULL, NULL}
};

int verbose = 0;

extern const char *__progname;

static struct option const longopts[] = {
	{"antialias", required_argument, 0, 'a'},
	{"autohint", required_argument, 0, 'u'},
	{"hinting", required_argument, 0, 'g'},
	{"hintstyle", required_argument, 0, 's'},
	{"rgba", required_argument, 0, 'r'},
	{"lcdfilter", required_argument, 0, 'l'},
	{"print", no_argument, 0, 'p'},
	{"verbose", no_argument, 0, 'v'},
	{"help", no_argument, 0, 'h'},
	{0, no_argument, 0, 0}
};

static void __attribute__ ((__noreturn__)) usage (FILE * stream, int status)
{
	fprintf (stream, "\
Usage: %s [--antialias (enable|disable)] [--autohint (enable|disable)] [--hinting (enable|disable)] [--hintstyle (disable|slight|medium|full)] [--rgba (disable|bgr|rgb|vbgr|vrgb)] [--lcdfilter (default|legacy|light|none)] [--help] [--verbose]\n\
fontconfig configuration utility.\n\
\n\
  --antialias   enabled/disabled antialiasing\n\
  --autohint    enabled/disabled autohinter\n\
  --hinting     enabled/disabled hinting\n\
  --hintstyle   set hinting style\n\
  --rgba        set sub-pixel rendering\n\
  --lcdfilter   set LCD filtering\n\
  --print       print settings and exit\n\
  --verbose     be verbose\n\
  --help        display this help and exit\n", __progname);
	exit (status);
}

static void link_file(const char *file)
{
	char src[PATH_MAX];
	char dst[PATH_MAX];

	if (!file)
		return;

	sprintf(src, AVAILDIR "%s", file);
	sprintf(dst, CONFDIR "%s", file);

	if (access(src, F_OK) != 0)
		sprintf(src, TEMPLATEDIR "%s", file);

	if (access(dst, F_OK) != 0) {
		if (symlink(src, dst) != 0) {
			if (verbose)
				fprintf(stderr, "failed link %s to %s\n", src, dst);
		}
	}
}

static void unlink_file(const char *file)
{
	char dst[PATH_MAX];

	if (!file)
		return;

	sprintf(dst, CONFDIR "%s", file);

	if (access(dst, F_OK) == 0) {
		if (unlink(dst) != 0) {
			if (verbose)
				fprintf(stderr, "failed unlink %s\n", dst);
		}
	}
}

static int getroot(void)
{
	if (getuid() != 0) {
		fprintf(stderr, "Sorry, you must be root to do this.\n");
		return 1;
	} else {
		return 0;
	}
}

static int get_conf(const char *file)
{
	char path[PATH_MAX];

	if (file) {
		sprintf(path, CONFDIR "%s", file);
		if (access(path, F_OK) == 0)
			return 0;
	} else {
		return 0;
	}

	return 1;
}

static void set_conf(struct FCStruct *S)
{
	struct FCStruct *s;

	for (s = S; s->name; s++) {
		unlink_file(s->file);
		if (strcmp(s->name, optarg) == 0)
			link_file(s->file);
	}
}

static void print_conf(struct FCStruct *S, const char *p)
{
	struct FCStruct *s;

	for (s = S; s->name; s++) {
		if (get_conf(s->file) == 0) {
			fprintf(stdout, "%s : %s\n", p, s->name);
			break;
		}
	}
}

static void get_settings(void)
{
	print_conf(Antialias, "antialiasing       ");
	print_conf(Autohint, "autohinter         ");
	print_conf(Hinting, "hinting            ");
	print_conf(Hintstyle, "hinting style      ");
	print_conf(RGBA, "sub-pixel rendering");
	print_conf(LCDFilter, "LCD filtering      ");
}

int main(int argc, char *argv[])
{
	int c;

	while ((c = getopt_long(argc, argv, "a:u:g:s:r:l:pvh", longopts, 0)) != -1)
		switch(c) {
			case 'a':
				if (getroot() == 0)
					set_conf(Antialias);
				else
					return 1;
				break;
			case 'u':
				if (getroot() == 0)
					set_conf(Autohint);
				else
					return 1;
				break;
			case 'g':
				if (getroot() == 0)
					set_conf(Hinting);
				else
					return 1;
				break;
			case 's':
				if (getroot() == 0)
					set_conf(Hintstyle);
				else
					return 1;
				break;
			case 'r':
				if (getroot() == 0)
					set_conf(RGBA);
				else
					return 1;
				break;
			case 'l':
				if (getroot() == 0)
					set_conf(LCDFilter);
				else
					return 1;
				break;
			case 'p':
				get_settings();
				return 0;
			case 'h':
				usage(stdout, EXIT_SUCCESS);
				break;
			case 'v':
				verbose = 1;
				break;
			default:
				usage(stderr, EXIT_FAILURE);
		}

	return 0;
}
